"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.deleteIds = exports.deleteId = exports.deleteSelector = exports.deleteHandler = exports.insertSelectorItem = exports.insertSelectorItems = exports.insertSelector = exports.insertHandler = void 0;
const client_1 = require("@deep-foundation/hasura/client");
const client_2 = require("./client");
const lodash_1 = __importDefault(require("lodash"));
const apolloClient = (0, client_1.generateApolloClient)({
    path: `${process.env.DEEPLINKS_HASURA_PATH}/v1/graphql`,
    ssl: !!+process.env.DEEPLINKS_HASURA_SSL,
    secret: process.env.DEEPLINKS_HASURA_SECRET,
});
const deep = new client_2.DeepClient({ apolloClient });
const insertHandler = (handleOperationTypeId, typeId, code, forceOwnerId, supportsId) => __awaiter(void 0, void 0, void 0, function* () {
    const syncTextFileTypeId = yield deep.id('@deep-foundation/core', 'SyncTextFile');
    const handlerJSFile = (yield deep.insert({
        type_id: syncTextFileTypeId,
    }, { name: 'INSERT_HANDLER_JS_FILE' })).data[0];
    const handlerJSFileValue = (yield deep.insert({ link_id: handlerJSFile === null || handlerJSFile === void 0 ? void 0 : handlerJSFile.id, value: code }, { table: 'strings' })).data[0];
    const handlerTypeId = yield deep.id('@deep-foundation/core', 'Handler');
    const handler = (yield deep.insert({
        from_id: supportsId || (yield deep.id('@deep-foundation/core', 'dockerSupportsJs')),
        type_id: handlerTypeId,
        to_id: handlerJSFile === null || handlerJSFile === void 0 ? void 0 : handlerJSFile.id,
    }, { name: 'INSERT_HANDLER' })).data[0];
    const containTypeId = yield deep.id('@deep-foundation/core', 'Contain');
    const ownerId = forceOwnerId || (yield deep.id('deep', 'admin'));
    const ownerContainHandler = (yield deep.insert({
        from_id: ownerId,
        type_id: containTypeId,
        to_id: handler === null || handler === void 0 ? void 0 : handler.id,
    }, { name: 'INSERT_ADMIN_CONTAIN_HANDLER' })).data[0];
    const handleOperation = (yield deep.insert({
        from_id: typeId,
        type_id: handleOperationTypeId,
        to_id: handler === null || handler === void 0 ? void 0 : handler.id,
    }, { name: 'INSERT_INSERT_HANDLER' })).data[0];
    return {
        handlerId: handler === null || handler === void 0 ? void 0 : handler.id,
        handleOperationId: handleOperation === null || handleOperation === void 0 ? void 0 : handleOperation.id,
        handlerJSFileId: handlerJSFile === null || handlerJSFile === void 0 ? void 0 : handlerJSFile.id,
        handlerJSFileValueId: handlerJSFileValue === null || handlerJSFileValue === void 0 ? void 0 : handlerJSFileValue.id,
        ownerContainHandlerId: ownerContainHandler === null || ownerContainHandler === void 0 ? void 0 : ownerContainHandler.id,
    };
});
exports.insertHandler = insertHandler;
function insertSelector() {
    return __awaiter(this, void 0, void 0, function* () {
        const typeTypeId = yield deep.id('@deep-foundation/core', 'Type');
        const { data: [{ id: ty0 }] } = yield deep.insert({
            type_id: typeTypeId,
        });
        const { data: [{ id: ty1 }] } = yield deep.insert({
            type_id: typeTypeId,
            from_id: ty0,
            to_id: ty0,
        });
        const treeIncludeDownTypeId = yield deep.id('@deep-foundation/core', 'TreeIncludeDown');
        const { data: [{ id: tr1, out: treeIncludes }] } = yield deep.insert({
            type_id: yield deep.id('@deep-foundation/core', 'Tree'),
            out: { data: [
                    {
                        type_id: treeIncludeDownTypeId,
                        to_id: ty1,
                    },
                    {
                        type_id: treeIncludeDownTypeId,
                        to_id: ty0,
                    },
                ] }
        }, {
            returning: `
      id
      out(where: { to_id: { _in: [${ty0}, ${ty1}] }, type_id: { _eq: ${treeIncludeDownTypeId} } }) {
        id
      }
    `,
        });
        const { data: [{ id: id0 }] } = yield deep.insert({
            type_id: ty0,
        });
        const selectorIncludeTypeId = yield deep.id('@deep-foundation/core', 'SelectorInclude');
        const selectorTreeTypeId = yield deep.id('@deep-foundation/core', 'SelectorTree');
        const { data: [{ id: s1, out: selectorData }] } = yield deep.insert({
            type_id: yield deep.id('@deep-foundation/core', 'Selector'),
            out: { data: [
                    {
                        type_id: selectorIncludeTypeId,
                        to_id: id0,
                        out: { data: {
                                type_id: selectorTreeTypeId,
                                to_id: tr1
                            } },
                    },
                ] }
        }, {
            returning: `
      id
      out(where: { to_id: { _eq: ${id0} }, type_id: { _eq: ${selectorIncludeTypeId} } }) {
        id
        out(where: { to_id: { _eq: ${tr1} }, type_id: { _eq: ${selectorTreeTypeId} } }) {
          id
        }
      }
    `,
        });
        return {
            nodeTypeId: ty0,
            linkTypeId: ty1,
            treeId: tr1,
            treeIncludesIds: treeIncludes.map(({ id }) => id),
            selectorId: s1,
            selectorIncludeId: selectorData[0].id,
            selectorTreeId: selectorData[0].out[0].id,
            rootId: id0,
        };
    });
}
exports.insertSelector = insertSelector;
;
function insertSelectorItems({ selectorId, nodeTypeId, linkTypeId, treeId, rootId }) {
    return __awaiter(this, void 0, void 0, function* () {
        const { data: [{ id: linkId1, to: { id: nodeId1 } }] } = yield deep.insert({
            type_id: linkTypeId,
            from_id: rootId,
            to: { data: {
                    type_id: nodeTypeId,
                } }
        }, { returning: 'id to { id }' });
        const { data: [{ id: linkId2, to: { id: nodeId2 } }] } = yield deep.insert({
            from_id: nodeId1,
            type_id: linkTypeId,
            to: { data: {
                    type_id: nodeTypeId,
                } }
        }, { returning: 'id to { id }' });
        return [
            { linkId: linkId1, nodeId: nodeId1 },
            { linkId: linkId2, nodeId: nodeId2 },
        ];
    });
}
exports.insertSelectorItems = insertSelectorItems;
;
function insertSelectorItem({ selectorId, nodeTypeId, linkTypeId, treeId, rootId }) {
    return __awaiter(this, void 0, void 0, function* () {
        const { data: [{ id: linkId }] } = yield deep.insert({
            type_id: linkTypeId,
            from_id: rootId,
            to: { data: {
                    type_id: nodeTypeId,
                } }
        }, { returning: 'id to { id }' });
        return { linkId };
    });
}
exports.insertSelectorItem = insertSelectorItem;
;
const deleteHandler = (handler) => __awaiter(void 0, void 0, void 0, function* () {
    const { handlerJSFileValueId } = handler, ids = __rest(handler, ["handlerJSFileValueId"]);
    const result = { links: [], strings: [] };
    const compact = { id: { _in: lodash_1.default.compact(Object.values(ids)) } };
    result.links.push(yield deep.delete(compact));
    return result;
});
exports.deleteHandler = deleteHandler;
const deleteSelector = (selector) => __awaiter(void 0, void 0, void 0, function* () {
    const { treeIncludesIds } = selector, withoutTreeIncluds = __rest(selector, ["treeIncludesIds"]);
    const ids = (lodash_1.default.concat(treeIncludesIds, Object.values(withoutTreeIncluds)));
    const compact = { id: { _in: lodash_1.default.compact(ids) } };
    yield deep.delete(compact);
});
exports.deleteSelector = deleteSelector;
function deleteId(id, options = { table: 'links' }) {
    return __awaiter(this, void 0, void 0, function* () {
        yield deleteIds([id], options);
    });
}
exports.deleteId = deleteId;
function deleteIds(ids, options = { table: 'links' }) {
    return __awaiter(this, void 0, void 0, function* () {
        const idsFiltered = ids === null || ids === void 0 ? void 0 : ids.filter(linkId => typeof linkId === 'number');
        if ((idsFiltered === null || idsFiltered === void 0 ? void 0 : idsFiltered.length) > 0) {
            try {
                return yield deep.delete(idsFiltered, options);
            }
            catch (e) {
                console.error(`Error deleting ids: ${idsFiltered.join(', ')}`, JSON.stringify(e, null, 2));
            }
        }
        else {
            return { data: [] };
        }
    });
}
exports.deleteIds = deleteIds;
//# sourceMappingURL=handlers.js.map