"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.delay = void 0;
const express_1 = __importDefault(require("express"));
const index_1 = __importDefault(require("./imports/router/index"));
const jwt_1 = __importDefault(require("./imports/router/jwt"));
const guest_1 = __importDefault(require("./imports/router/guest"));
const packager_1 = __importDefault(require("./imports/router/packager"));
const authorization_1 = __importDefault(require("./imports/router/authorization"));
const axios_1 = __importDefault(require("axios"));
const http_1 = __importDefault(require("http"));
const http_proxy_middleware_1 = require("http-proxy-middleware");
const graphql_playground_middleware_express_1 = __importDefault(require("graphql-playground-middleware-express"));
const moesif_nodejs_1 = __importDefault(require("moesif-nodejs"));
const debug_1 = __importDefault(require("debug"));
const wait_on_1 = __importDefault(require("wait-on"));
const client_1 = require("@deep-foundation/hasura/client");
const client_2 = require("./imports/client");
const graphql_tag_1 = __importDefault(require("graphql-tag"));
const links_1 = require("./imports/router/links");
const lodash_1 = __importDefault(require("lodash"));
const DEEPLINKS_HASURA_PATH = process.env.DEEPLINKS_HASURA_PATH || 'localhost:8080';
const DEEPLINKS_HASURA_SSL = process.env.DEEPLINKS_HASURA_SSL || 0;
const DEEPLINKS_HASURA_SECRET = process.env.DEEPLINKS_HASURA_SECRET || 'myadminsecretkey';
const MOESIF_TOKEN = process.env.MOESIF_TOKEN || '';
const DEEPLINKS_PUBLIC_URL = process.env.DEEPLINKS_PUBLIC_URL || '';
const debug = (0, debug_1.default)('deeplinks');
const log = debug.extend('log');
const error = debug.extend('error');
const namespaces = debug_1.default.disable();
debug_1.default.enable(`${namespaces ? `${namespaces},` : ``}${error.namespace}`);
const delay = (time) => new Promise(res => setTimeout(() => res(null), time));
exports.delay = delay;
const app = (0, express_1.default)();
const httpServer = http_1.default.createServer(app);
app.get('/gql', (0, graphql_playground_middleware_express_1.default)({
    tabs: [{
            endpoint: `${DEEPLINKS_PUBLIC_URL}/gql`,
            query: `query MyQuery {
      links(limit: 1) {
        id
      }
    }`,
            headers: {
                Authorization: 'Bearer TOKEN',
            },
        }],
}));
app.use('/gql', (0, http_proxy_middleware_1.createProxyMiddleware)({
    target: `http${DEEPLINKS_HASURA_SSL === '1' ? 's' : ''}://${DEEPLINKS_HASURA_PATH}`,
    changeOrigin: true,
    ws: true,
    pathRewrite: {
        "/gql": "/v1/graphql",
    },
}));
app.use(['/v1', '/v1alpha1', '/v2', '/console'], (0, http_proxy_middleware_1.createProxyMiddleware)({
    target: `http${DEEPLINKS_HASURA_SSL === '1' ? 's' : ''}://${DEEPLINKS_HASURA_PATH}`,
    changeOrigin: true,
    ws: true,
}));
if (MOESIF_TOKEN) {
    const moesifMiddleware = (0, moesif_nodejs_1.default)({ applicationId: MOESIF_TOKEN });
    app.use(moesifMiddleware);
    moesifMiddleware.startCaptureOutgoing();
}
app.use(express_1.default.json());
app.use('/', index_1.default);
const start = () => __awaiter(void 0, void 0, void 0, function* () {
    const jwtServer = (0, jwt_1.default)(httpServer);
    const authorizationServer = (0, authorization_1.default)(httpServer);
    const guestServer = (0, guest_1.default)(httpServer);
    const packagerServer = (0, packager_1.default)(httpServer);
    yield jwtServer.start();
    yield guestServer.start();
    yield authorizationServer.start();
    yield packagerServer.start();
    jwtServer.applyMiddleware({ path: '/api/jwt', app });
    guestServer.applyMiddleware({ path: '/api/guest', app });
    authorizationServer.applyMiddleware({ path: '/api/authorization', app });
    packagerServer.applyMiddleware({ path: '/api/packager', app });
    yield new Promise(resolve => httpServer.listen({ port: process.env.PORT }, resolve));
    log(`Hello bugfixers! Listening ${process.env.PORT} port`);
    try {
        yield (0, wait_on_1.default)({ resources: [`http${DEEPLINKS_HASURA_SSL === '1' ? 's' : ''}-get://${DEEPLINKS_HASURA_PATH}/console`] });
        yield (0, axios_1.default)({
            method: 'post',
            url: `http${DEEPLINKS_HASURA_SSL === '1' ? 's' : ''}://${DEEPLINKS_HASURA_PATH}/v1/metadata`,
            headers: { 'x-hasura-admin-secret': DEEPLINKS_HASURA_SECRET, 'Content-Type': 'application/json' },
            data: { type: 'reload_metadata', args: {} }
        }).then(() => {
            log('hasura metadata reloaded');
        }, () => {
            error('hasura metadata broken');
        });
    }
    catch (e) {
        error(e);
    }
});
start();
const client = (0, client_1.generateApolloClient)({
    path: `${process.env.DEEPLINKS_HASURA_PATH}/v1/graphql`,
    ssl: !!+process.env.DEEPLINKS_HASURA_SSL,
    secret: process.env.DEEPLINKS_HASURA_SECRET,
});
const deep = new client_2.DeepClient({
    apolloClient: client,
});
const routesDebug = (0, debug_1.default)('deeplinks').extend('eh').extend('routes');
const routesDebugLog = routesDebug.extend('log');
let currentServers = {};
let currentPorts = {};
let busy = false;
let portTypeId = 0;
const toJSON = (data) => JSON.stringify(data, Object.getOwnPropertyNames(data), 2);
const handleRoutes = () => __awaiter(void 0, void 0, void 0, function* () {
    var _a, _b, _c, _d, _e, _f, _g, _h;
    if (busy)
        return;
    busy = true;
    try {
        portTypeId = yield deep.id('@deep-foundation/core', 'Port');
        const handleRouteTypeId = yield deep.id('@deep-foundation/core', 'HandleRoute');
        const routerStringUseTypeId = yield deep.id('@deep-foundation/core', 'RouterStringUse');
        const routerListeningTypeId = yield deep.id('@deep-foundation/core', 'RouterListening');
        const routesResult = yield client.query({
            query: (0, graphql_tag_1.default) `
        query {
          ports: links(where: {
            type_id: { _eq: "${portTypeId}" }
          }) {
            id
            port: value
            routerListening: in(where: {
              type_id: { _eq: "${routerListeningTypeId}" }
            }) {
              id
              router: from {
                id
                routerStringUse: in(where: {
                  type_id: { _eq: "${routerStringUseTypeId}" }
                }) {
                  id
                  routeString: value
                  route: from {
                    id
                    handleRoute: out(where: {
                      type_id: { _eq: "${handleRouteTypeId}" }
                    }) {
                      id
                      handler: to {
                        id
                        supports: from {
                          id
                          isolation: from {
                            id
                            image: value
                          }
                        }
                        file: to {
                          id
                          code: value
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      `, variables: {}
        });
        const portsResult = routesResult.data.ports;
        routesDebugLog('portsResult', JSON.stringify(portsResult, null, 2));
        const ports = {};
        for (const port of portsResult) {
            const portValue = (_a = port === null || port === void 0 ? void 0 : port.port) === null || _a === void 0 ? void 0 : _a.value;
            ports[portValue] = port;
        }
        routesDebugLog('ports', JSON.stringify(ports, null, 2));
        const updatedOrAddedPorts = [];
        for (const key in currentPorts) {
            if (currentPorts.hasOwnProperty(key)) {
                if (ports.hasOwnProperty(key)) {
                    if (!lodash_1.default.isEqual(currentPorts[key], ports[key])) {
                        currentPorts[key] = ports[key];
                        updatedOrAddedPorts.push(ports[key]);
                    }
                    else {
                    }
                }
                else {
                    if (currentServers.hasOwnProperty(key)) {
                        const element = currentServers[key];
                        element.close();
                        delete currentServers[key];
                    }
                    delete currentPorts[key];
                }
            }
        }
        for (const key in ports) {
            if (ports.hasOwnProperty(key)) {
                if (!currentPorts.hasOwnProperty(key)) {
                    currentPorts[key] = ports[key];
                    updatedOrAddedPorts.push(ports[key]);
                }
            }
        }
        routesDebugLog('updatedOrAddedPorts', JSON.stringify(updatedOrAddedPorts, null, 2));
        routesDebugLog('currentPorts', JSON.stringify(currentPorts, null, 2));
        const imageContainers = {};
        updatedOrAddedPorts.forEach(port => {
            port.routerListening.forEach(routerListening => {
                var _a;
                (_a = routerListening === null || routerListening === void 0 ? void 0 : routerListening.router) === null || _a === void 0 ? void 0 : _a.routerStringUse.forEach(routerStringUse => {
                    var _a;
                    (_a = routerStringUse === null || routerStringUse === void 0 ? void 0 : routerStringUse.route) === null || _a === void 0 ? void 0 : _a.handleRoute.forEach(handleRoute => {
                        var _a, _b, _c, _d;
                        imageContainers[(_d = (_c = (_b = (_a = handleRoute === null || handleRoute === void 0 ? void 0 : handleRoute.handler) === null || _a === void 0 ? void 0 : _a.supports) === null || _b === void 0 ? void 0 : _b.isolation) === null || _c === void 0 ? void 0 : _c.image) === null || _d === void 0 ? void 0 : _d.value] = {};
                    });
                });
            });
        });
        const imageList = Object.keys(imageContainers);
        routesDebugLog('imageList', imageList);
        for (const image of imageList) {
            routesDebugLog(`preparing container ${image}`);
            imageContainers[image] = yield links_1.containerController.newContainer({
                handler: image,
                forceRestart: true,
                publish: +links_1.DOCKER ? false : true,
                code: '',
                jwt: '',
                data: {}
            });
        }
        for (const port of updatedOrAddedPorts) {
            const portValue = (_b = port === null || port === void 0 ? void 0 : port.port) === null || _b === void 0 ? void 0 : _b.value;
            if (currentServers.hasOwnProperty(portValue)) {
                currentServers[portValue].close();
            }
            if (port.routerListening.length > 0) {
                routesDebugLog(`listening on port ${portValue}`);
                const portServer = (0, express_1.default)();
                currentServers[portValue] = http_1.default.createServer({ maxHeaderSize: 10 * 1024 * 1024 * 1024 }, portServer).listen(portValue);
                for (const routerListening of port.routerListening) {
                    const router = routerListening === null || routerListening === void 0 ? void 0 : routerListening.router;
                    for (const routerStringUse of router.routerStringUse) {
                        const routeString = (_c = routerStringUse === null || routerStringUse === void 0 ? void 0 : routerStringUse.routeString) === null || _c === void 0 ? void 0 : _c.value;
                        routesDebugLog(`route string ${routeString}`);
                        const route = routerStringUse === null || routerStringUse === void 0 ? void 0 : routerStringUse.route;
                        for (const handleRoute of route.handleRoute) {
                            const handler = handleRoute === null || handleRoute === void 0 ? void 0 : handleRoute.handler;
                            const handlerId = handler === null || handler === void 0 ? void 0 : handler.id;
                            routesDebugLog(`handler id ${handlerId}`);
                            const jwt = yield (0, links_1.getJwt)(handlerId, routesDebugLog);
                            routesDebugLog(`jwt ${jwt}`);
                            const image = (_f = (_e = (_d = handler === null || handler === void 0 ? void 0 : handler.supports) === null || _d === void 0 ? void 0 : _d.isolation) === null || _e === void 0 ? void 0 : _e.image) === null || _f === void 0 ? void 0 : _f.value;
                            routesDebugLog(`image`, image);
                            const container = imageContainers[image];
                            routesDebugLog(`container`, JSON.stringify(container, null, 2));
                            const code = (_h = (_g = handler === null || handler === void 0 ? void 0 : handler.file) === null || _g === void 0 ? void 0 : _g.code) === null || _h === void 0 ? void 0 : _h.value;
                            routesDebugLog(`code ${code}`);
                            console.log('container', container);
                            const proxy = (0, http_proxy_middleware_1.createProxyMiddleware)({
                                target: `http://${container.host}:${container.port}`,
                                changeOrigin: true,
                                ws: true,
                                pathRewrite: {
                                    [routeString]: "/http-call",
                                },
                                onProxyReq: (proxyReq, req, res) => {
                                    console.log('deeplinks request');
                                    console.log('req.method', req.method);
                                    console.log('req.body', req.body);
                                    proxyReq.setHeader('deep-call-options', encodeURI(JSON.stringify({
                                        jwt,
                                        code,
                                        data: {},
                                    })));
                                },
                                onProxyRes: (proxyRes, req, res) => {
                                    proxyRes.on('data', function (data) {
                                        return __awaiter(this, void 0, void 0, function* () {
                                            try {
                                                data = data.toString('utf-8');
                                                console.log('data', data);
                                                if (data.startsWith('{')) {
                                                    data = JSON.parse(data);
                                                    if (data.hasOwnProperty('rejected')) {
                                                        console.log('rejected', data.rejected);
                                                        const handlingErrorTypeId = yield deep.id('@deep-foundation/core', 'HandlingError');
                                                        console.log('handlingErrorTypeId', handlingErrorTypeId);
                                                        const insertResult = yield deep.insert({
                                                            type_id: handlingErrorTypeId,
                                                            object: { data: { value: data.rejected } },
                                                            out: { data: [
                                                                    {
                                                                        type_id: yield deep.id('@deep-foundation/core', 'HandlingErrorReason'),
                                                                        to_id: route.id
                                                                    },
                                                                    {
                                                                        type_id: yield deep.id('@deep-foundation/core', 'HandlingErrorReason'),
                                                                        to_id: handleRoute.id
                                                                    }
                                                                ] },
                                                        }, {
                                                            name: 'INSERT_HANDLING_ERROR',
                                                        });
                                                    }
                                                }
                                            }
                                            catch (e) {
                                                console.log('deeplinks response error', e);
                                            }
                                        });
                                    });
                                }
                            });
                            portServer.use(routeString, proxy);
                        }
                    }
                }
            }
        }
    }
    catch (e) {
        routesDebugLog(toJSON(e));
    }
    busy = false;
});
const startRouteHandling = () => __awaiter(void 0, void 0, void 0, function* () {
    setInterval(handleRoutes, 5000);
});
startRouteHandling();
//# sourceMappingURL=index.js.map